/*
   FILE:							       	
	trace/instr_macros.h
				       	
   DESCRIPTION:							       	
	Macros containing the instrumentation points inside the
	code of the RT-Linux scheduler. Each system event type
	has its own macro.
					       				
   Copyright (C) 2001-04
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        December 2004

*/
#ifndef __TRACE_INSTR_MACROS_H
#define __TRACE_INSTR_MACROS_H


// 1) Context switch:
#ifdef CONFIG_OC_PTRACE_SYSEV_CSWITCH
#define POSIX_TRACE_SYSEV_CSWITCH(_newtask) \
       { \
	  if (TRACE_SYSTEM_EVENT(POSIX_TRACE_CSWITCH)) { \
	    posix_trace_sysevent_data data; \
	    data.index    = 0; \
	    data.task     = _newtask->posix_trace_id; \
	    data.priority = _newtask->sched_param.sched_priority; \
	    posix_trace_system_event(POSIX_TRACE_CSWITCH, &data, \
				     sizeof(data), 0); \
	  } \
       }
#else 
#define POSIX_TRACE_SYSEV_CSWITCH(_ntask)
#endif


// 2) State change for tasks:
#ifdef CONFIG_OC_PTRACE_SYSEV_TSKSTATE
#define POSIX_TRACE_SYSEV_TSKSTATE(_newstate,_tsk) \
       { \
          if (TRACE_SYSTEM_EVENT(POSIX_TRACE_TASK_NEW_STATE)) { \
            posix_trace_sysevent_data data; \
            data.index    = _newstate; \
            data.task     = _tsk->posix_trace_id; \
	    posix_trace_system_event(POSIX_TRACE_TASK_NEW_STATE, \
                                     &data, sizeof(data),0); \
	  } \
       }

#else
#define POSIX_TRACE_SYSEV_TSKSTATE(_newstate,_tsk)
#endif


// 3) RT-Linux scheduler starts:
#ifdef CONFIG_OC_PTRACE_SYSEV_SCHSTRT
#define POSIX_TRACE_SYSEV_SCHSTRT() \
	{ \
	  if (TRACE_SYSTEM_EVENT(POSIX_TRACE_SCHED_START)) { \
	    posix_trace_sysevent_data data; \
	    data.index = posix_trace_timer_interrupt; \
	    posix_trace_system_event(POSIX_TRACE_SCHED_START, \
				     &data, sizeof(data), 0); \
	    posix_trace_timer_interrupt = 0; \
	  } \
	}
#else 
#define POSIX_TRACE_SYSEV_SCHSTRT()
#endif


// 4) RT-Linux scheduler ends:
#ifdef CONFIG_OC_PTRACE_SYSEV_SCHEND
#define POSIX_TRACE_SYSEV_SCHEND() \
      { \
	if (TRACE_SYSTEM_EVENT(POSIX_TRACE_SCHED_END)) { \
	  /* TODO: probar a enviar un NULL en "data" */ \
	  posix_trace_sysevent_data data; \
	  posix_trace_system_event(POSIX_TRACE_SCHED_END, \
				   &data, sizeof(data), 0); \
	} \
      } 
#else
#define POSIX_TRACE_SYSEV_SCHEND()
#endif


// 5) Start of processing a system call:
#ifdef  CONFIG_OC_PTRACE_SYSEV_SYSCALL
#define POSIX_TRACE_SYSEV_SYSCALL(_sysc,_tsk) \
      { \
	if (TRACE_SYSTEM_EVENT(POSIX_TRACE_SYSCALL)) { \
	  int calling_address; \
	  posix_trace_sysevent_data data; \
	  data.index   = _sysc; \
	  data.task    = (_tsk)->posix_trace_id; \
	  posix_trace_get_invoking_address(calling_address); \
	  posix_trace_system_event(POSIX_TRACE_SYSCALL,&data, \
                                   sizeof(data), calling_address); \
	} \
      }

// The clock_nanosleep syscall is a bit different:
#define POSIX_TRACE_SYSEV_SYSCALL_NANOSLP(_tsk,_rqtp) \
      { \
	if (TRACE_SYSTEM_EVENT(POSIX_TRACE_SYSCALL)) { \
	  int calling_address; \
	  posix_trace_sysevent_data data; \
	  data.index   = SYSCALL_CLOCK_NANOSLEEP; \
	  data.task    = (_tsk)->posix_trace_id; \
	  data.other_data.delay_until.tv_sec  = _rqtp->tv_sec; \
	  data.other_data.delay_until.tv_nsec = _rqtp->tv_nsec; \
	  posix_trace_get_invoking_address(calling_address); \
	  posix_trace_system_event(POSIX_TRACE_SYSCALL,&data, \
				   sizeof(data), calling_address); \
	} \
      }
#else 
#define POSIX_TRACE_SYSEV_SYSCALL(_sys,_tsk)
#define POSIX_TRACE_SYSEV_SYSCALL_NANOSLP(_tsk,_rqtp)
#endif


// 6) End of processing a system call:
#ifdef CONFIG_OC_PTRACE_SYSEV_SYSEND
#define POSIX_TRACE_SYSEV_SYSEND(_sys,_tsk,_errcode) \
      { \
	if (TRACE_SYSTEM_EVENT(POSIX_TRACE_SYSCALL_END)) { \
	  int calling_address; \
	  posix_trace_sysevent_data data; \
	  data.index   = _sys; \
	  data.task    = (_tsk)->posix_trace_id; \
	  data.return_value = _errcode; \
	  posix_trace_get_invoking_address(calling_address); \
	  posix_trace_system_event(POSIX_TRACE_SYSCALL_END, &data, \
				   sizeof(data),calling_address); \
	  } \
	}
#else
#define POSIX_TRACE_SYSEV_SYSEND(_sys,_tsk,_errcode)
#endif


// 7) Mutex lock:
#ifdef CONFIG_OC_PTRACE_SYSEV_MUTLOCK
#define POSIX_TRACE_SYSEV_MUTLOCK(_result,_tsk,_mtx) \
      {  \
	if (TRACE_SYSTEM_EVENT(POSIX_TRACE_MUTEX_LOCK)) { \
	  posix_trace_sysevent_data data; \
	  data.index = _result; \
	  data.task  = (_tsk)->posix_trace_id; \
	  data.other_data.mutex_data.mutex = (_mtx)->posix_trace_id; \
	  posix_trace_system_event(POSIX_TRACE_MUTEX_LOCK,&data, \
				   sizeof(data),0); \
	} \
      }
#else
#define POSIX_TRACE_SYSEV_MUTLOCK(_result,_tsk,_mtx)
#endif


// 8) Mutex unlock:
#ifdef CONFIG_OC_PTRACE_SYSEV_MUTULOCK
#define POSIX_TRACE_SYSEV_MUTULOCK(_tsk,_mtx) \
      { \
	if (TRACE_SYSTEM_EVENT(POSIX_TRACE_MUTEX_UNLOCK)) { \
	  posix_trace_sysevent_data data; \
	  data.task  = (_tsk)->posix_trace_id; \
	  data.other_data.mutex_data.mutex = (_mtx)->posix_trace_id; \
	  posix_trace_system_event(POSIX_TRACE_MUTEX_UNLOCK,&data, \
				   sizeof(data),0); \
	} \
      }
#else
#define POSIX_TRACE_SYSEV_MUTULOCK()
#endif


// 9) Condition variable wait:
#ifdef CONFIG_OC_PTRACE_SYSEV_CONDWAIT
#define POSIX_TRACE_SYSEV_CONDWAIT()
#else
#define POSIX_TRACE_SYSEV_CONDWAIT()
#endif

// 10) Condition variable broadcast:
#ifdef CONFIG_OC_PTRACE_SYSEV_CONDBCAST
#define POSIX_TRACE_SYSEV_CONDBCAST() 
#else
#define POSIX_TRACE_SYSEV_CONDBCAST()
#endif



#endif // #ifndef __TRACE_INSTR_MACROS_H












