/*
   FILE:
						       	
	test_ptrace_api2_module.c
								       	
   DESCRIPTION:							       	

        Test 2 for POSIX Tracing API in RT-Linux.

	Basically, this test revises the tracing API referred to
	streams, except the actual tracing and retrieval of events.

	In particular, the following API functions are tested:

	  + Creation of a stream without log with some particular 
	    attributes
	  + Creation of new (user) event types
	  + Rewind and traversal of the list of event types
	  + Manipulation of sets of event types
	  + Get/Set filters from/to streams
	  + Retrieval of stream's attributes and status info
	  + Shut a stream down

					       				
   Copyright (C) 2004
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        September 2004

*/

#include <rtl.h>
#include <time.h>
#include <trace.h>
#include <asm/errno.h>


MODULE_LICENSE("GPL");
MODULE_AUTHOR("Andres Terrasa");
MODULE_DESCRIPTION("API Test 2 for POSIX Tracing");


#define TEST_NAME "ptrace_test_api2"


#define MY_EVENT_DATA_SIZE   (size_t)    64
#define MY_STREAM_SIZE       (size_t)  4096



int init_module(void) {

  trace_attr_t   attr;
  int            error;
  int            unavailable;
  int            user_event_found;
  char           event_name[TRACE_EVENT_NAME_MAX];
  trace_id_t       trid;
  trace_event_id_t event_int_id, event_string_id, new_event_string_id;
  trace_event_id_t first_event_id, event_id;
#ifdef _POSIX_TRACE_EVENT_FILTER
  trace_event_set_t  set1, set2;
  int                ismember;
  int                ismember2;
#endif
#ifdef _POSIX_TRACE_ONLINE
  size_t             size;
  int                full_policy;
  struct posix_trace_status_info 
    statusinfo;
#endif
  
  // Create the trace attribute:
  if ((error = posix_trace_attr_init(&attr)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_attr_init returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  
  if ((error = posix_trace_attr_setstreamfullpolicy(&attr, 
						    POSIX_TRACE_UNTIL_FULL)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_attr_setstreamfullpolicy returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }  

  if ((error = posix_trace_attr_setmaxdatasize(&attr, MY_EVENT_DATA_SIZE)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_attr_setmaxdatasize returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_attr_setstreamsize(&attr, MY_STREAM_SIZE)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_attr_setmaxdatasize returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Create the stream (without log)
  if ((error = posix_trace_create(0, &attr, &trid)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_create returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  // Create two new event types, with a different function each:
  if ((error = posix_trace_eventid_open ("user event string", 
					 &event_string_id)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_eventid_open returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


#ifdef _POSIX_TRACE_EVENT_FILTER
  if ((error = posix_trace_trid_eventid_open (trid, "user event integer", 
					      &event_int_id)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_trid_eventid_open returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
#endif

  // Get name for first one and check the name is ok
  if ((error = posix_trace_eventid_get_name (trid, event_string_id, 
					     event_name)) != 0) {

    rtl_printf("%s: [BROK] :posix_trace_trid_eventid_get_name returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if (strcmp(event_name, "user event string")) {
    rtl_printf("%s: [BROK] :Event name set and get different\n", TEST_NAME);
    return -1;
  }

  // Define the same event again and test if both event ids are equal:
  if ((error = posix_trace_eventid_open ("user event string", 
					 &new_event_string_id)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_eventid_open returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if (! posix_trace_eventid_equal (trid, event_string_id, new_event_string_id)) {
    rtl_printf("%s: [BROK] :posix_trace_eventid_equal finds two events are not equal\n", 
	       TEST_NAME);
    return -1;
  }

  /* Rewind the list of event types and get all of them.
     Check that the user event created above is within the list.
     When done, rewind it again and check the list starts with 
     the first one: 
  */
  if ((error = posix_trace_eventtypelist_rewind(trid)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventtypelist_rewind returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_eventtypelist_getnext_id (trid, &event_id, 
						     &unavailable)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_eventtypelist_getnext_id returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
  
  first_event_id = event_id;
  user_event_found = 0;

  while (! unavailable) {

    user_event_found |= posix_trace_eventid_equal (trid, event_string_id, 
						  event_id);

    if ((error = posix_trace_eventtypelist_getnext_id (trid, &event_id, 
						       &unavailable)) != 0) {
      rtl_printf("%s: [BROK] :posix_trace_eventtypelist_getnext_id returned %d\n", 
		 TEST_NAME, error);
      return -1;
    }
  }

  if (! user_event_found) {
      rtl_printf("%s: [BROK] :User event is NOT on the list of event types\n", 
		 TEST_NAME);
      return -1;
  }

  if ((error = posix_trace_eventtypelist_rewind(trid)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventtypelist_rewind returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_eventtypelist_getnext_id (trid, &event_id, 
						     &unavailable)) != 0) {
    rtl_printf("%s: [BROK] :posix_trace_eventtypelist_getnext_id returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }
 
  if (! posix_trace_eventid_equal (trid, event_id, first_event_id)) {
      rtl_printf("%s: [BROK] :After rewind twice, first event in the list is different\n", 
		 TEST_NAME);
      return -1;
  }


#ifdef _POSIX_TRACE_EVENT_FILTER

  /* 
     Event set manipulation: 
       + create set1 full, check the user event is there
       + create set2 empty, check user event is not there
       + remove user event from set1, check now it's not there
       + add user event to set2, check now it's thre
  */

  if ((error = posix_trace_eventset_fill(&set1, POSIX_TRACE_ALL_EVENTS)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_fill returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_eventset_empty(&set2)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_empty returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  if ((error = posix_trace_eventset_ismember(event_string_id, &set1, 
					     &ismember )) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_ismember returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if (! ismember) {
      rtl_printf("%s: [BROK] :event id should be a filter member, but it's NOT\n", 
		 TEST_NAME);
      return -1;
  }

  if ((error = posix_trace_eventset_ismember(event_string_id, &set2, 
					     &ismember )) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_ismember returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if (ismember) {
      rtl_printf("%s: [BROK] :event id should NOT be a filter member, but it is\n", 
		 TEST_NAME);
      return -1;
  }

  if ((error = posix_trace_eventset_del(event_string_id, &set1)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_del returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_eventset_ismember(event_string_id, &set1, 
					     &ismember )) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_ismember returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if (ismember) {
      rtl_printf("%s: [BROK] :event id should NOT be a filter member, but it is\n", 
		 TEST_NAME);
      return -1;
  }

  if ((error = posix_trace_eventset_add(event_string_id, &set2)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_add returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_eventset_ismember(event_string_id, &set2, 
					     &ismember )) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_eventset_ismember returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if (! ismember) {
      rtl_printf("%s: [BROK] :event id should be a filter member, but it's NOT\n", 
		 TEST_NAME);
      return -1;
  }
  

  /* 
     Event filter set/get 
       + Set a filter with all system event types and no user event types
        (for example) to the stream
       + Get the stream's current filter
       + Check the two filters contain the same event types.

  */

  posix_trace_eventset_fill(&set1, POSIX_TRACE_SYSTEM_EVENTS);

  if ((error = posix_trace_set_filter(trid, &set1, 
				      POSIX_TRACE_SET_EVENTSET)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if ((error = posix_trace_get_filter(trid, &set2)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_get_filter returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  posix_trace_eventtypelist_rewind(trid);
  posix_trace_eventtypelist_getnext_id (trid, &event_id, &unavailable);

  while (! unavailable) {

    posix_trace_eventset_ismember(event_id, &set1, &ismember);
    posix_trace_eventset_ismember(event_id, &set2, &ismember2);

    if (ismember != ismember2) {
      rtl_printf("%s: [BROK] :Set filter different from get filter\n", 
	       TEST_NAME);
      return -1;
    }
    
    posix_trace_eventtypelist_getnext_id (trid, &event_id, &unavailable);

  }


#endif


#ifdef _POSIX_TRACE_ONLINE
  /*
    Retrieve the stream's attributes and status
    + Attributes should be equal to the ones set in the creation (check some)
    + Status has to be initial, since no tracing has been done
  */

  if ((error = posix_trace_get_attr(trid, &attr)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_get_attr returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  posix_trace_attr_getstreamfullpolicy(&attr, &full_policy);
  if (full_policy != POSIX_TRACE_UNTIL_FULL) {
    rtl_printf("%s: [BROK] :Full policy attribute different from expected\n", 
	       TEST_NAME);
    return -1;
  }

  posix_trace_attr_getmaxdatasize(&attr, &size);
  if (size != MY_EVENT_DATA_SIZE) {
    rtl_printf("%s: [BROK] :Event data size different from expected\n", 
	       TEST_NAME);
    return -1;
  }

  posix_trace_attr_getstreamsize(&attr, &size);
  if (size != MY_STREAM_SIZE) {
    rtl_printf("%s: [BROK] :Stream size different from expected\n", 
	       TEST_NAME);
    return -1;
  }

  if ((error = posix_trace_get_status(trid, &statusinfo)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_get_status returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }

  if (statusinfo.posix_stream_status != POSIX_TRACE_SUSPENDED ||
      statusinfo.posix_stream_full_status != POSIX_TRACE_NOT_FULL ||
      statusinfo.posix_stream_overrun_status != POSIX_TRACE_NO_OVERRUN) {
    rtl_printf("%s: [BROK] :Stream's status wrong. Not initial values\n", 
	       TEST_NAME);
    return -1;
  }


#endif

  /* Shut the stream down */

  if ((error = posix_trace_shutdown(trid)) != 0) { 
    rtl_printf("%s: [BROK] :posix_trace_shutdown returned %d\n", 
	       TEST_NAME, error);
    return -1;
  }


  
  
  /* No errors. All tests were fine. Return 0. */
  rtl_printf("%s: PASS :Stream-related API works OK\n", TEST_NAME);  
  return 0;

}



void cleanup_module(void) {

}
