/*
   FILE:							       	
	rtl_posixtrace.c					       
								       	
   DESCRIPTION:							       	
	Posix trace implementation in RT-Linux (v 3.0)
					       				
   Copyright (C) 2001
	Ana Garcia Fornes
	Andres Terrasa Barrena
	Ignacio Paches Giner
	Agustin Espinosa Minguet

  This is free software;  you can  redistribute it  and/or modify it under
  terms of the  GNU General Public License as published  by the Free Soft-
  ware  Foundation;  either version 2,  or (at your option) any later ver-
  sion.  It is distributed in the hope that it will be useful, but WITH-
  OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
  for  more details.  You should have  received  a copy of the GNU General
  Public License;  see file COPYING.  If not, write to  the Free Software 
  Foundation,  59 Temple Place - Suite 330,  Boston, MA 02111-1307, USA.

  LAST REVISION:
        August 2001

*/

#include <stdlib.h>
#include <unistd.h>

#include <rtl_time.h>
#include <rtl_sched.h>
#include <rtl_fifo.h>
#include <mbuff.h>
#include <time.h>

#include <trace.h>
#include <trace/comm.h>

#ifdef _POSIX_TRACE_CORE

//#define _TRACE_DEBUG

/* 
   MODULE DEFINITIONS
*/

#ifdef _POSIX_TRACE_STREAM_SIZE
#define TRACE_STREAM_MEM_SIZE   (size_t) _POSIX_TRACE_STREAM_SIZE
#else
#define TRACE_STREAM_MEM_SIZE   0x100000 // Default value for stream_min_size
#endif

#define TRACE_STREAM_MAGIC      0xaafd
#define TRACE_LOG_MAGIC         0xaadf
#define TRACE_ATTR_MAGIC        0xabef
#define NO_MAGIC                0x0

#define MAX_MBUFF_NAME          32

#define MAX_WAITING_TASKS       4

#define TRACE_FIFO_SIZE         4096

#define TRACE_FULL_POLICY_DEFAULT  -POSIX_TRACE_LOOP 

#define FLUSHING_MAX_DEV_SIZE   40000
#define FLUSHING_BURST_INTERVAL 40000000LL

// Types of active streams: WLOG (With LOG) and WOLOG (WithOut LOG)
// (pre-recorded streams are not supported in RT-Linux, only in Linux)
#define TRACE_STREAM_WOLOG      1
#define TRACE_STREAM_WLOG       2

// This implementation only supports streams which can store at least this
// amount of events:
#define MIN_NUMBER_OF_EVENTS    6   

#define OVERFLOW                1
#define NO_OVERFLOW             0


/* 
   MODULE TYPES 
*/


// Structures to build the stack of threads suspended in the stream
typedef struct wait_thread_stack {
  pthread_t       list[MAX_WAITING_TASKS];
  int             top;
} trace_wait_stack_t;


// Each stream is controlled by means of this structure
typedef struct trace_stream_header { 
  int                      magic;
  int                      type;
  pid_t                    posix_pid;
  trace_attr_t             attr; 
  posix_trace_status_info  status;

  event_type_info_t        * event_types; 
  int                      current_event;

  trace_event_set_t        filter;

  char                     mbuff_name[MAX_MBUFF_NAME];
  size_t                   event_size;
  int                      max_events;
  int                      n_events;

  stream_event_t           * events;
  stream_event_t           * in;
  stream_event_t           * out;
  stream_event_t           * last_event;

#ifdef _POSIX_TRACE_ONLINE
  int                      overflow_condition;
#endif
#ifdef _POSIX_TRACE_TOLOG
  int                      log_fd;
#endif

  void (*do_posix_trace_event) (struct trace_stream_header *stream, 
			       trace_event_id_t event_id, 
			       const void *data_ptr,      
			       size_t data_len,
			       int calling_address);

  trace_wait_stack_t       wait_stack;
 
  struct trace_stream_header * next; // To build the list of active streams
} trace_info_t;

#define NO_STREAM  (trace_info_t *) 0
#define EOL        (trace_info_t *) 0
#define NONEXT     (trace_info_t *) -1



/*
   MODULE GLOBAL VARIABLES
*/

// Array of streams
static trace_info_t        rtl_trace_streams[TRACE_SYS_MAX];

// Dynamic list of active (created) streams (initially empty)
static trace_info_t      *rtl_active_stream_list = EOL;

// Maximum system event size (this is set at the initialization stage)
static size_t max_sysevent_data_size = (size_t) 0;

// Boolean to detect if we're under an overflow condition:
// (TODO: isn't this a per-stream condition??? STUDY IT!)
//#ifdef _POSIX_TRACE_ONLINE
//static int  overflow_condition = 0;
//#endif

// Global filter of system event, initially all events are filtered
int global_system_events_filter =~0;


// Array of event types defined for the target (shared by all streams)
// (NOTE: Remember that, in RTLinux, there is only one target) 
// Any change in the system event type definition should also be
// reflected here
static event_type_info_t   rtl_trace_event_types[TRACE_EVENT_MAX] = {
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {-1                           , ""},  {-1                           , ""},
  {POSIX_TRACE_UNNAMED_USEREVENT, "posix_trace_unnamed_userevent"},
  {POSIX_TRACE_ERROR            , "posix_trace_error"},
  {POSIX_TRACE_START            , "posix_trace_start"},
  {POSIX_TRACE_STOP             , "posix_trace_stop"},
  {POSIX_TRACE_OVERFLOW         , "posix_trace_overflow"},
  {POSIX_TRACE_RESUME           , "posix_trace_resume"},
  {POSIX_TRACE_FILTER           , "posix_trace_filter"},
  {POSIX_TRACE_CSWITCH          , "posix_trace_cswitch" }, 
  {POSIX_TRACE_TASK_NEW_STATE   , "posix_trace_task_new_state" },
  {POSIX_TRACE_SCHED_START      , "posix_trace_sched_start" },
  {POSIX_TRACE_SCHED_END        , "posix_trace_sched_end" }, 
  {POSIX_TRACE_SYSCALL          , "posix_trace_syscall" }, 
  {POSIX_TRACE_SYSCALL_END      , "posix_trace_syscall_end" }, 
  {POSIX_TRACE_MUTEX_LOCK       , "posix_trace_mutex_lock"},
  {POSIX_TRACE_MUTEX_UNLOCK     , "posix_trace_mutex_unlock"},
  {POSIX_TRACE_COND_WAIT        , "posix_trace_cond_wait"},
  {POSIX_TRACE_COND_BROADCAST   , "posix_trace_cond_broadcast"},
  {POSIX_TRACE_CSWITCH_TOLINUX  , "posix_trace_cswitch_tolinux" }, 
  {POSIX_TRACE_CSWITCH_TOTASK   , "posix_trace_cswitch_totask" }, 
  {-1                           , "" },
  {-1                           , "" }, {-1                           , ""},
  {-1                           , "" }, {-1                           , ""},
  {-1                           , "" }, {-1                           , ""},
  {-1                           , "" }, {-1                           , ""},
  {-1                           , "" }, {-1                           , ""},
  {-1                           , "" }, {-1                           , ""}
};






/* 
   MACRO DEFINITIONS
*/

#define MIN(a,b)                  a < b ? a : b

#define MAX(a,b)                  a < b ? b : a


#define TRACE_ATTR_IS_VALID(attr) ( \
                                     attr->magic == TRACE_ATTR_MAGIC \
                                  )


#define TRACE_FIND_STREAM(trid)   ( \
                                     trid >= 0 && trid < TRACE_SYS_MAX && \
                                     rtl_trace_streams[trid].magic == \
                                        TRACE_STREAM_MAGIC ? \
                                     & rtl_trace_streams[trid] : \
                                     NO_STREAM \
                                  )


#define TRACE_EVENT_IS_VALID(ev) ( \
                                     ev >= 0 && ev < TRACE_EVENT_MAX \
                                 )


#define TRACE_EVENT_IS_DEFINED(event) ( \
                                     (event).eventid >= 0 \
                                 )


#define TRACE_EVENT_IS_USER(ev) \
                                 ( \
                                     ev < TRACE_USER_EVENT_MAX \
                                 )


#define TRACE_EVENT_IS_FILTERED(ev,set)\
                                 ( \
                                    ev < TRACE_USER_EVENT_MAX? \
				    ((set).user_set & 1<<(ev))  :  \
                                    ((set).system_set & 1<<(ev-32))  \
                                 )

#define TRACE_USER_EVENT_IS_FILTERED(ev,set)\
                                 ( \
                                    constant_test_bit(event_id, \
				                     (volatile void *) &(set.user_set)) \
                                 )

//                                     (set).user_set & 1<<(ev)


#define TRACE_STREAM_IS_FULL(stream) \
                                  (  \
                                     stream->n_events == stream->max_events \
                                  )


#define TRACE_STREAM_IS_RUNNING(stream) \
                                  (  \
                                     stream->status.posix_stream_status == \
                                     POSIX_TRACE_RUNNING \
                                  )


#define TRACE_STREAM_IS_EMPTY(stream) \
                                  (  \
                                     stream->n_events == 0 \
                                  )


#define TRACE_STREAM_INCREMENT_PTR(stream,position) \
                                  (  \
                                     position == stream->last_event? \
                                     (stream_event_t *) stream->events : \
                                     (stream_event_t *)  \
                                      ((int) position + (int) stream->event_size) \
                                  )


#define TRACE_STACK_IS_EMPTY(stack) \
                                  (  \
                                     stack.top == -1 \
				  )


#define TRACE_STACK_IS_FULL(stack) \
                                  (  \
                                     stack.top == (MAX_WAITING_TASKS-1) \
				  )


#define TRACE_STACK_PUSH(stack,task) \
                                  { \
                                     stack.list[++stack.top] = task; \
                                  };

#define TRACE_STACK_POP(stack,task)  \
                                  { \
                                    if (stack.top == -1) \
                                       { task = 0; } \
                                    else { task = stack.list[stack.top]; \
                                           stack.top--; } \
                                  };


#define TRACE_TASK_PRIORITY(task) ( \
                                    task->sched_param.sched_priority \
                                  )



/*
  AUXILIAR FUNCTION TO BUILD THE GLOBAL FILTER

  The global filter contains the events types which are filtered in
  all the running streams.

*/
static inline void build_global_filter(void) {
  // Called with interrupts disabled
  trace_info_t *stream;

  global_system_events_filter =~0;

  for (stream = rtl_active_stream_list; stream != EOL; stream = stream->next) {
    if (TRACE_STREAM_IS_RUNNING(stream)) {
      global_system_events_filter &= stream->filter.system_set;
    }
  }
}
  


/* 
   ADVANCED FUNCTION DECLARATIONS
*/
#ifdef _POSIX_TRACE_LINUXTRACE
int rtf_target_handler(unsigned int fifo);
#endif

static inline 
 void _do_posix_trace_event_loop (trace_info_t *stream,
			     trace_event_id_t event_id, 
 /* Implementation */        const void *data_ptr,      
 /*     below      */        size_t data_len,
                             int calling_address);

static inline 
 void _do_posix_trace_event_untilfull (trace_info_t *stream,
			     trace_event_id_t event_id, 
 /* Implementation */        const void *data_ptr,      
 /*     below      */        size_t data_len,
                             int calling_address);


/* 
   ---------------------------------------------------------------------
   24.3. IMPLEMENTATION OF THE TRACE FUNCTIONS
   ---------------------------------------------------------------------
*/

//
// 24.3.2. Attribute initialization and destruction
//
int posix_trace_attr_init (trace_attr_t * attr) {
  hrtime_t clock_res;

  // Initialize the attribute with default values:
  attr->magic                    = TRACE_ATTR_MAGIC;
  strncpy(attr->generation_version, 
	 "RTL POSIX Trace 2.0 (c) 2003 Terrasa,Garcia-Fornes,Espinosa",
	 sizeof(attr->generation_version));
  strncpy(attr->trace_name, "", sizeof(attr->trace_name));
  clock_res                      = gethrtimeres();
  attr->clock_resolution         = timespec_from_ns(clock_res);
  attr->stream_full_policy       = TRACE_FULL_POLICY_DEFAULT;

#ifdef _POSIX_TRACE_EVENT_SIZE
  attr->data_size = MAX((size_t) max_sysevent_data_size,
			(size_t) _POSIX_TRACE_EVENT_SIZE);
#else
  attr->data_size                = (size_t) max_sysevent_data_size;
#endif
  attr->stream_min_size          = (size_t) TRACE_STREAM_MEM_SIZE;
  attr->max_data_size            = attr->data_size;
  attr->header_size              = sizeof(stream_event_t);

#ifdef _POSIX_TRACE_INHERITANCE
  attr->inheritance              = POSIX_TRACE_CLOSE_FOR_CHILD;
#endif
#ifdef _POSIX_TRACE_TOLOG
  /*
    The following default value for "log_full_policy" is not the POSIX
    one, but this implementation only supports the POSIX_TRACE_APPEND
    log full policy (on which log_max_size does not matter).
  */
  attr->log_max_size            = -1;
  attr->log_full_policy         = POSIX_TRACE_APPEND; 
#endif

   return 0;
}





int posix_trace_attr_destroy (trace_attr_t *attr) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  // Destroy the attribute
  attr->magic = NO_MAGIC;
  return 0;
}




//
// 24.3.3. Retrieve and set information about a trace stream
//
int posix_trace_attr_getgenversion (const trace_attr_t *attr, 
				    char *genversion) {
  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  strncpy(genversion, attr->generation_version, TRACE_NAME_MAX);
  return 0;
}





int posix_trace_attr_getname (const trace_attr_t *attr,
			      char *tracename) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  strncpy(tracename, attr->trace_name, TRACE_NAME_MAX);
  return 0;
}





int posix_trace_attr_setname (trace_attr_t *attr, 
			      const char *tracename) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  strncpy(attr->trace_name, tracename, sizeof(attr->trace_name));
  attr->trace_name[TRACE_NAME_MAX-1] = '\0';
  return 0;
}





int posix_trace_attr_getcreatetime (const trace_attr_t *attr, 
				    struct timespec *createtime) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *createtime = attr->creation_time;
  return 0;
}





int posix_trace_attr_getclockres (const trace_attr_t *attr,
				  struct timespec *resolution) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *resolution = attr->clock_resolution;
  return 0;
}




//
// 24.3.4. Retrieve and set the behavior of a trace stream
//
int posix_trace_attr_getstreamfullpolicy (const trace_attr_t *attr, 
					  int *streampolicy) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *streampolicy  = abs(attr->stream_full_policy);
  return 0;
}





int posix_trace_attr_setstreamfullpolicy (trace_attr_t *attr, 
					  int streampolicy) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  if (streampolicy != POSIX_TRACE_LOOP 
      && streampolicy != POSIX_TRACE_UNTIL_FULL 
#ifdef _POSIX_TRACE_TOLOG
      && streampolicy != POSIX_TRACE_FLUSH 
#endif
     ) {
    return -EINVAL;
  }

  attr->stream_full_policy = streampolicy;
  return 0;
}




#ifdef _POSIX_TRACE_TOLOG

int posix_trace_attr_getlogfullpolicy (const trace_attr_t *attr, 
				       int *logpolicy) {
  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *logpolicy = attr->log_full_policy;
  return 0;
}




int posix_trace_attr_setlogfullpolicy (trace_attr_t *attr, 
				       int logpolicy) {
  
  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  if (logpolicy != POSIX_TRACE_APPEND) {
    // This implementation *only* supports POSIX_TRACE_APPEND
    return -EINVAL;
  }

  return 0;
}

#endif






//
// 24.3.5. Retrieve and set trace stream size attributes
//
// Notes: 
//       This implementation uses an array of fixed-size elements
//       to store events into a stream. Thus, the maximum amount of 
//       space required to store an event does not depend on this event
//       being a user or a system event.
//       The space occupied by any stored event is the event header
//       (stream_event_t) plus the maximum allowable data space 
//       (attr->max_data_size), even if the event has a smaller data size. 
//
//       The max_data_size value can be increased or decreased by the
//       user as needed; this value will be used to truncate
//       any *user* event with greater data size. However, the **real** 
//       maximum amount of space required to store a single event's data  
//       (attr->data_size) cannot be less than the amount of space required
//       to store the a system event's data (max_sysevent_data_size).

int posix_trace_attr_getmaxusereventsize (const trace_attr_t *attr, 
					   size_t data_len, 
					   size_t *eventsize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *eventsize= attr->header_size + attr->data_size - sizeof(void *);
  return 0;
}





int posix_trace_attr_getmaxsystemeventsize (const trace_attr_t *attr,
					    size_t *eventsize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *eventsize= attr->header_size + attr->data_size - sizeof(void *);
  return 0;
}





int posix_trace_attr_getmaxdatasize (const trace_attr_t *attr, 
				     size_t *maxdatasize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

  *maxdatasize = attr->max_data_size;
  return 0;
}





int posix_trace_attr_setmaxdatasize (trace_attr_t *attr, 
				     size_t maxdatasize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

#ifdef _POSIX_TRACE_EVENT_SIZE
  if (maxdatasize <= (size_t) _POSIX_TRACE_EVENT_SIZE) {
    attr->max_data_size = maxdatasize;
    attr->data_size     = MAX(maxdatasize, max_sysevent_data_size);
  } else {
    return -EINVAL;
  }
#else
  attr->max_data_size = maxdatasize;
  attr->data_size     = MAX(maxdatasize, max_sysevent_data_size);
#endif
  return 0;
}





int posix_trace_attr_getstreamsize (const trace_attr_t *attr, 
				    size_t *streamsize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }
 
  *streamsize = attr->stream_min_size;
  return 0;
}





int posix_trace_attr_setstreamsize (trace_attr_t *attr, 
				    size_t streamsize) {

  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }

#ifdef _POSIX_TRACE_STREAM_SIZE
  if (streamsize <= (size_t) _POSIX_TRACE_STREAM_SIZE) {
    attr->stream_min_size = streamsize;
  } else {
    return -EINVAL;
  }
#else
  attr->stream_min_size = streamsize;
#endif
  return 0;
}





#ifdef _POSIX_TRACE_TOLOG

int posix_trace_attr_getlogsize (const trace_attr_t *attr, 
				 size_t *logsize) {
  // Error checking
  if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }
  
  *logsize = attr->log_max_size;
  return 0;
}




int posix_trace_attr_setlogsize(trace_attr_t *attr, 
				size_t logsize) {

  /*
    Since the only log full policy supported is POSIX_TRACE_APPEND, on
    which log size is not relevant, the implementation doesn't allow
    the application to change this value.
  */
  return -EINVAL;
}

#endif





//
// 24.3.6. Trace stream initialization, flush and shutdown
// 

int posix_trace_eventset_empty (trace_event_set_t *set); // Advanced definition

#ifdef _POSIX_TRACE_ONLINE
int posix_trace_create (pid_t pid,
			const trace_attr_t *attr,
			trace_id_t *trid) {
  rtl_irqstate_t flags;
  int            trid_id;
  size_t         stream_size;
  int            max_events;
  size_t         event_size;
  void           *mbuff_addr;
  trace_info_t   *stream;
  char           mbuff_name[MAX_MBUFF_NAME];
  trace_attr_t   default_attr;

  // Cancellation test
  pthread_testcancel();

  // Error checking
  if (attr == (trace_attr_t *) NULL) {
    // Create a default attr object
    posix_trace_attr_init(&default_attr);
    attr = &default_attr;
  } else if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }
#ifdef _POSIX_TRACE_TOLOG
  if (attr->stream_full_policy == POSIX_TRACE_FLUSH) {
    return -EINVAL;
  }
#endif
  if (pid != 0) { 
    return -ESRCH;
  }
   
  // Search for the first stream which isn't being used
  for (trid_id = 0; 
       trid_id < TRACE_SYS_MAX 
	 && rtl_trace_streams[trid_id].magic == TRACE_STREAM_MAGIC; 
       trid_id++) /* void loop */ ;
  
  if (trid_id >= TRACE_SYS_MAX) {
    return -EAGAIN;
  } 
  
  /*
    This implementation doesn't put a lower bound on the
    attr->stream_min_size value. However, this value must be big
    enough to successfully store (at least) a minimum,
    implementation-dependent, amount of events (MIN_NUMBER_OF_EVENTS).
    Below this value, the tracing system may not work properly.
  */
  stream_size = attr->stream_min_size;
  event_size  = sizeof(stream_event_t) - sizeof(void *) + attr->data_size;
  max_events  = (int) (stream_size / event_size);

#ifdef _TRACE_DEBUG
  rtl_printf("(ptrace_create) stream_size=%d event_size=%d max_events=%d\n", 
	     stream_size, event_size, max_events);
#endif

  if (max_events < MIN_NUMBER_OF_EVENTS) {
   return -EINVAL;
  }

  // Open the mbuff related to the stream
  sprintf(mbuff_name, "stream %d mbuff", trid_id);
  if ((mbuff_addr = mbuff_alloc(mbuff_name, stream_size)) == NULL) { 
    return -ENOMEM;
  }

  // OK. No errors. Initialize the stream:
  *trid  = trid_id;
  stream = &rtl_trace_streams[*trid];
  
  rtl_no_interrupts(flags);
  // Stream's general data:
  stream->magic      = TRACE_STREAM_MAGIC;
  stream->type       = TRACE_STREAM_WOLOG;
  stream->posix_pid  = pid;
  strncpy(stream->mbuff_name, mbuff_name, sizeof(stream->mbuff_name));
  stream->wait_stack.top = -1;

  // Stream's attr object:
  memcpy(&(stream->attr), attr, sizeof(trace_attr_t));
  if (stream->attr.stream_full_policy == TRACE_FULL_POLICY_DEFAULT) {
    stream->attr.stream_full_policy = POSIX_TRACE_LOOP;
  }
  clock_gettime(CLOCK_RTL_SCHED,   &(stream->attr.creation_time));

  // Stream's initial status:
  stream->status.posix_stream_status         = POSIX_TRACE_SUSPENDED;
  stream->status.posix_stream_full_status    = POSIX_TRACE_NOT_FULL;
  stream->status.posix_stream_overrun_status = POSIX_TRACE_NO_OVERRUN;
#ifdef _POSIX_TRACE_TOLOG
  stream->status.posix_stream_flush_status   = POSIX_TRACE_NOT_FLUSHING;
  stream->status.posix_stream_flush_error    = 0;
  stream->status.posix_log_overrun_status    = POSIX_TRACE_NO_OVERRUN;
  stream->status.posix_log_full_status       = POSIX_TRACE_NOT_FULL;
#endif
  stream->overflow_condition                 = 0;

  // Stream's event types and filter set:
  stream->event_types   = rtl_trace_event_types;
  stream->current_event = 0;
  posix_trace_eventset_empty(&stream->filter);
  // Stream's array of stored events:
  stream->events        = (stream_event_t *) mbuff_addr;
  stream->in            = stream->events;
  stream->out           = stream->events;
  stream->n_events      = 0;
  stream->event_size    = event_size;
  stream->max_events    = max_events; 
  stream->last_event    = (stream_event_t *) 
                             ((int) stream->events + 
                             ((int) (stream->event_size * 
                                    (stream->max_events-1))));

  // Stream's trace function pointer:
  if (stream->attr.stream_full_policy == POSIX_TRACE_UNTIL_FULL) {
    stream->do_posix_trace_event = _do_posix_trace_event_untilfull;
  } else {
    stream->do_posix_trace_event = _do_posix_trace_event_loop;
  }
    

#ifdef _TRACE_DEBUG
  rtl_printf("(ptrace_create) START=%x  SIZE=%d  MAX=%d  LAST=%x\n", 
	     stream->events, (int) stream->event_size, 
	     (int) stream->max_events, stream->last_event);
#endif

  // Insert the stream into the list of active streams:
  if (stream->next == NONEXT) {
    stream->next = rtl_active_stream_list;
    rtl_active_stream_list = stream;
  }
  rtl_restore_interrupts(flags);  

  return 0;
}
#endif




static inline int _do_posix_trace_start (trace_info_t *stream);
                 /* Implementation below */ 



#ifdef _POSIX_TRACE_TOLOG

int posix_trace_create_withlog (pid_t pid,
				const trace_attr_t *attr,
				int file_desc,
				trace_id_t *trid) {
  rtl_irqstate_t flags;
  int            trid_id;
  size_t         stream_size;
  int            max_events;
  size_t         event_size;
  void           *mbuff_addr;
  trace_info_t   *stream;
  char           mbuff_name[MAX_MBUFF_NAME];
  trace_attr_t   default_attr;
  flush_header_t flush_info;
  int            ret;

  // Cancellation test
  pthread_testcancel();

  // Error checking
  if (attr == (trace_attr_t *) NULL) {
    // Create a default attr object
    posix_trace_attr_init(&default_attr);
    attr = &default_attr;
  } else if (! TRACE_ATTR_IS_VALID(attr)) {
    return -EINVAL;
  }
  if (attr->stream_full_policy == POSIX_TRACE_FLUSH) {
    return -EINVAL;
  }
  if (pid != 0) { 
    return -ESRCH;
  }

  // Search for the first stream which isn't being used
  for (trid_id = 0; 
       trid_id < TRACE_SYS_MAX 
	 && rtl_trace_streams[trid_id].magic == TRACE_STREAM_MAGIC; 
       trid_id++) /* void loop */ ;
  
  if (trid_id >= TRACE_SYS_MAX) {
    return -EAGAIN;
  } 
  
  /*
    As this is a stream with log, POSIX establishes that the file
    descriptor has to be valid (opened and writable) at this time.
    Otherwise, EBADF is returned
  */
  flush_info.type = FLUSH_LOG_STARTING;
  flush_info.add_bytes = 0;
  ret = write(file_desc, &flush_info, sizeof(flush_info));

  if (ret < sizeof(flush_info)) {
    return -EBADF;
  }

  /*
    This implementation doesn't put a lower bound on the
    attr->stream_min_size value. However, this value must be big
    enough to successfully store (at least) a minimum,
    implementation-dependent, amount of events (MIN_NUMBER_OF_EVENTS).
    Below this value, the tracing system may not work properly.
  */
  stream_size = attr->stream_min_size;
  event_size  = sizeof(stream_event_t) - sizeof(void *) + attr->data_size;
  max_events  = (int) (stream_size / event_size);

#ifdef _TRACE_DEBUG
  rtl_printf("(ptrace_create_withlog) stream_size=%d event_size=%d max_events=%d\n", 
	     stream_size, event_size, max_events);
#endif

  if (max_events < MIN_NUMBER_OF_EVENTS) {
   return -EINVAL;
  }

  // Open the mbuff related to the stream
  sprintf(mbuff_name, "stream %d mbuff", trid_id);
  if ((mbuff_addr = mbuff_alloc(mbuff_name, stream_size)) == NULL) { 
    return -ENOMEM;
  }

  // OK. No errors. Initialize the stream:
  *trid  = trid_id;
  stream = &rtl_trace_streams[*trid];
  
  rtl_no_interrupts(flags);
  // Stream's general data:
  stream->magic      = TRACE_STREAM_MAGIC;
  stream->type       = TRACE_STREAM_WLOG;
  stream->log_fd     = file_desc;
  stream->posix_pid  = pid;
  strncpy(stream->mbuff_name, mbuff_name, sizeof(stream->mbuff_name));
  stream->wait_stack.top = -1;

  // Stream's attr object:
  memcpy(&(stream->attr), attr, sizeof(trace_attr_t));
  if (stream->attr.stream_full_policy == TRACE_FULL_POLICY_DEFAULT) {
    stream->attr.stream_full_policy = POSIX_TRACE_UNTIL_FULL;
  }
  clock_gettime(CLOCK_RTL_SCHED,   &(stream->attr.creation_time));

  // Stream's initial status:
  stream->status.posix_stream_status         = POSIX_TRACE_SUSPENDED;
  stream->status.posix_stream_full_status    = POSIX_TRACE_NOT_FULL;
  stream->status.posix_stream_overrun_status = POSIX_TRACE_NO_OVERRUN;
  stream->status.posix_stream_flush_status   = POSIX_TRACE_NOT_FLUSHING;
  stream->status.posix_stream_flush_error    = 0;
  stream->status.posix_log_overrun_status    = POSIX_TRACE_NO_OVERRUN;
  stream->status.posix_log_full_status       = POSIX_TRACE_NOT_FULL;

  // Stream's event types and filter set:
  stream->event_types   = rtl_trace_event_types;
  stream->current_event = 0;
  posix_trace_eventset_empty(&stream->filter);

  // Stream's array of stored events:
  stream->events        = (stream_event_t *) mbuff_addr;
  stream->in            = stream->events;
  stream->out           = stream->events;
  stream->n_events      = 0;
  stream->event_size    = event_size;
  stream->max_events    = max_events; 
  stream->last_event    = (stream_event_t *) 
    ((int) stream->events + 
     ((int) (stream->event_size * 
	     (stream->max_events-1))));

  // Stream's trace function pointer:
  if (stream->attr.stream_full_policy == POSIX_TRACE_UNTIL_FULL) {
    stream->do_posix_trace_event = _do_posix_trace_event_untilfull;
  } else {
    stream->do_posix_trace_event = _do_posix_trace_event_loop;
  }

#ifdef _TRACE_DEBUG
  rtl_printf("(ptrace_create_withlog) START=%x SIZE=%d MAX=%d LAST=%x\n", 
	     stream->events, (int) stream->event_size, 
	     (int) stream->max_events, stream->last_event);
#endif

  // Insert the stream into the list of active streams:
  if (stream->next == NONEXT) {
    stream->next = rtl_active_stream_list;
    rtl_active_stream_list = stream;
  }
  rtl_restore_interrupts(flags);  

  return 0;
}




static inline int _do_posix_trace_flush (trace_info_t *stream) {
  rtl_irqstate_t    flags;
  stream_event_t   *current_in;
  stream_event_t   *copy_out;
  int               n_events_flushed;
  int               n_events_to_flush = 0;
  int               written;
  flush_header_t    flush_info;
  int               bytes_flushed = 0;
  size_t            event_header_size;
  int               error = 0;
  hrtime_t          timeout;

  rtl_no_interrupts(flags);
  if (stream->n_events == 0) {
    rtl_restore_interrupts(flags);
    return 0;
  }

  // Initialize local variables which control the flushing
  current_in         = stream->in;
  copy_out           = stream->out;
  n_events_to_flush  = stream->n_events;
  n_events_flushed   = 0;
  event_header_size  = sizeof(stream_event_t) - sizeof(void *);
  stream->status.posix_stream_flush_status = POSIX_TRACE_FLUSHING;
  rtl_restore_interrupts(flags);

#ifdef _TRACE_DEBUG
  rtl_printf("(ptrace_flush): Begin # events in stream %d (in:%x  out:%x)\n", 
	     n_events_to_flush, current_in, copy_out);
#endif


  // Flushing main loop:
  do {

    flush_info.type      = FLUSH_LOG_EVENT;
    flush_info.add_bytes = event_header_size + copy_out->data_length;

    if ((written=write(stream->log_fd, &flush_info, sizeof(flush_info))) < 0) {
      error = -EFBIG; 
      break;
    }

    copy_out->timestamp = tsc2nsecs(copy_out->timestamp);
    if ((written=write(stream->log_fd, copy_out, 
	               event_header_size + copy_out->data_length)) < 0) {
      error = -EFBIG;
      break;
    }

    bytes_flushed += (sizeof(flush_info) + event_header_size 
		      + copy_out->data_length);
    n_events_flushed++;
    copy_out = TRACE_STREAM_INCREMENT_PTR(stream,copy_out);

    /* 
       TODO (ATB): 
       Change flushing protocol based on bytes_flushed to simply error
       codes and retrials: if written less than expected, then sleep
       for a while and then *resend* the same event. This is the only
       thing to do since we don't know the space left on device.
    */
    if (bytes_flushed >= FLUSHING_MAX_DEV_SIZE) {
#ifdef _TRACE_DEBUG
	rtl_printf("   (flush) Flushed: %d of %d (%d bytes) (out:%x)\n",
		   n_events_flushed, n_events_to_flush, bytes_flushed, 
		   copy_out);
#endif
      bytes_flushed = 0;
      // Sleep for a while so Linux process can void the device:
      if (pthread_self() != pthread_linux()) {
	timeout = gethrtime() + (hrtime_t) FLUSHING_BURST_INTERVAL;
	timeout = __rtl_fix_timeout_for_clock(CLOCK_REALTIME, timeout);
	__rtl_setup_timeout(pthread_self(), timeout);
	RTL_MARK_SUSPENDED(pthread_self());
	rtl_schedule(); 
      }
    }

  } while (n_events_flushed < n_events_to_flush);  
  
  
  rtl_no_interrupts(flags);
  // Now change stream status:
  stream->out = copy_out;
  stream->n_events = stream->n_events - n_events_flushed;

  // If the stream was full and the flushing has succeeded, then it is
  // not full again and we may have to start it running again:
  if (stream->status.posix_stream_full_status == POSIX_TRACE_FULL 
      && ! TRACE_STREAM_IS_FULL(stream)) {
    stream->status.posix_stream_full_status = POSIX_TRACE_NOT_FULL;
    if (stream->attr.stream_full_policy == POSIX_TRACE_UNTIL_FULL) {
      _do_posix_trace_start(stream);
    }
  }	

  // Done with the flushing:
  stream->status.posix_stream_flush_status = POSIX_TRACE_NOT_FLUSHING;
  stream->status.posix_stream_flush_error  = error;
  rtl_restore_interrupts(flags);

#ifdef _TRACE_DEBUG
  rtl_printf("(ptrace_flush): End # events left %d (in:%x out:%x) Error:%d\n", 
	     stream->n_events, current_in, copy_out, error);
#endif

  return error;

} 





int posix_trace_flush (trace_id_t trid) {
  trace_info_t     *stream;

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM 
      || stream->type != TRACE_STREAM_WLOG) {
    return -EINVAL;
  }

  return _do_posix_trace_flush(stream);
}

#endif





extern int posix_trace_shutdown (trace_id_t trid) {

  trace_info_t  *stream;
  rtl_irqstate_t flags;
  pthread_t      task;
#ifdef _POSIX_TRACE_TOLOG
  hrtime_t       timeout;
#endif

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  // First stop the stream (which rebuilds the global filter)
  posix_trace_stop(trid);

  // Remove the stream from list of active streams (no more tracing there)
  rtl_no_interrupts(flags);
  // Remove the stream from the active list:
  if (rtl_active_stream_list == stream) {
    rtl_active_stream_list = stream->next; 
  } else {
    trace_info_t *prev,*next;
    for (prev = rtl_active_stream_list, next = prev->next;
	 next != stream && next != EOL;
	 prev = next, next = next->next) /* void */;
    if (next != EOL) {
      prev->next = next->next;
    }
  }
  rtl_restore_interrupts(flags);

  // Some rt-tasks may be waiting on this stream. If so, awaken them
  // (with an error):
  if (stream->type == TRACE_STREAM_WOLOG) {
    while (! TRACE_STACK_IS_EMPTY(stream->wait_stack)) {
      TRACE_STACK_POP(stream->wait_stack, task);
      if (task) {
	pthread_kill(task, RTL_SIGNAL_WAKEUP);
      }
    }
  }


#ifdef _POSIX_TRACE_TOLOG
  /*
    If this is a stream with log, flush all its events before losing
    them and then write the definitive information about the stream's
    attributes, event types and status.
  */
  if (stream->type == TRACE_STREAM_WLOG) {
    flush_header_t    flush_info;

#ifdef _TRACE_DEBUG
    rtl_printf("(ptrace_shutdown): Flushing the stream\n");
#endif
  _do_posix_trace_flush(stream);

  // After flushing and before sending the final information, make
  // sure there is enough space left on the flushing device:
  if (pthread_self() != pthread_linux()) {
    timeout = gethrtime() + (hrtime_t) FLUSHING_BURST_INTERVAL;
    timeout = __rtl_fix_timeout_for_clock(CLOCK_REALTIME, timeout);
    __rtl_setup_timeout(pthread_self(), timeout);
    RTL_MARK_SUSPENDED(pthread_self());
    rtl_schedule(); 
  }
    
#ifdef _TRACE_DEBUG
  rtl_printf("(ptrace_shutdown): Sending attibutes, event types and status\n");
#endif

    flush_info.type = FLUSH_LOG_HEADER;
    flush_info.add_bytes = sizeof(trace_attr_t) 
      + sizeof(rtl_trace_event_types)
      + sizeof(posix_trace_status_info);
    write(stream->log_fd, &flush_info, sizeof(flush_info));

    write(stream->log_fd, &stream->attr, sizeof(trace_attr_t));
    write(stream->log_fd, rtl_trace_event_types, 
	  sizeof(event_type_info_t) * TRACE_EVENT_MAX);
    write(stream->log_fd, &stream->status, sizeof(posix_trace_status_info));
   
    flush_info.type = FLUSH_LOG_DONE;
    flush_info.add_bytes = 0;
    write(stream->log_fd, &flush_info, sizeof(flush_info));
    
#ifdef _TRACE_DEBUG
    rtl_printf("(ptrace_shutdown): Done with the log\n");
#endif
  }    
#endif
  
  // Finally release the stream resources:
  rtl_no_interrupts(flags);
  stream->next = NONEXT;
  stream->magic = NO_MAGIC;
  mbuff_free(stream->mbuff_name, stream->events); 
  rtl_restore_interrupts(flags);

  // Cancellation test
  pthread_testcancel();
  return 0;
}





//
// 24.3.7. Clear trace stream and trace log
//
extern int posix_trace_clear (trace_id_t trid) {

  trace_info_t  *stream;
  rtl_irqstate_t flags;  

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();

  // Void the stream of events
  rtl_no_interrupts(flags);
  stream->in            = stream->events;
  stream->out           = stream->events;
  stream->n_events    = 0;
  stream->status.posix_stream_full_status = POSIX_TRACE_NOT_FULL;
  rtl_restore_interrupts(flags);
  return 0;
}





//
// 24.3.8. Manipulate trace event type identifier
//

// Note: This private function does the job of
// "posix_trace_trid_eventid_open()" and "posix_trace_eventid_open()"
// below

static inline int _do_posix_trace_eventid_open(event_type_info_t *events,
					       const char *event_name,
					       trace_event_id_t *event) {

  int           eventid = 0;
  int           found   = 0;
  rtl_irqstate_t flags;  

  // Error checking
  if (strlen(event_name) > TRACE_EVENT_NAME_MAX) {
    return -ENAMETOOLONG;
  }

  // First look for the event name in the array of event ids:

  eventid = 0;
  found   = 0;
  while (! found && eventid < TRACE_EVENT_MAX) {

    found = (TRACE_EVENT_IS_DEFINED(events[eventid]) 
	     &&
	     ! strcmp(events[eventid].name, event_name));

    if (! found) {
      eventid ++;
    }
  }

  // If found, return the type id, else define the new user even type:
  if (found) {
    *event = eventid; 
  } else {
    for (eventid = 0; 
	 eventid < TRACE_USER_EVENT_MAX 
	   && TRACE_EVENT_IS_DEFINED(events[eventid]);
	 eventid ++) /* void */ ;

    if (eventid < TRACE_USER_EVENT_MAX) {
      // New user event
      *event = eventid;
      rtl_no_interrupts(flags);
      events[eventid].eventid = eventid;
      strncpy(events[eventid].name, event_name, sizeof(events[eventid].name));
      rtl_restore_interrupts(flags);
    } else {
      // No room for a new user event type; return the generic user event
      *event = POSIX_TRACE_UNNAMED_USEREVENT;
    }
  }
  return 0;
}




#ifdef _POSIX_TRACE_EVENT_FILTER
int posix_trace_trid_eventid_open (trace_id_t trid,
				   const char *event_name,
				   trace_event_id_t *event) {

  trace_info_t  *stream;

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  return _do_posix_trace_eventid_open(stream->event_types, event_name, event);
}
#endif 





int posix_trace_eventid_get_name (trace_id_t trid,
				  trace_event_id_t event,
				  char *event_name ) {

  trace_info_t  *stream;

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  if ( ! TRACE_EVENT_IS_DEFINED(stream->event_types[event])) {
    return -EINVAL;
  }

  strncpy(event_name, stream->event_types[event].name, TRACE_EVENT_NAME_MAX);
  return 0;
}





int posix_trace_eventid_equal (trace_id_t trid,
			       trace_event_id_t event1,
			       trace_event_id_t event2) {

  trace_info_t  *stream;

  // Note: According to the standard, there's NO error checking in
  // this function

  // If the stream exists and both events are defined there, then return 
  // whether they are equal or not
  if ((stream = TRACE_FIND_STREAM(trid)) != NO_STREAM) {
    if (TRACE_EVENT_IS_DEFINED(stream->event_types[event1]) 
	&& TRACE_EVENT_IS_DEFINED(stream->event_types[event2])) {
       return (event1 == event2);
    }
  }
  // In any other case, we consider these events different
  return 0;
}





//
// 24.3.9. Iterate over a mapping of trace event types
//
int posix_trace_eventtypelist_getnext_id (trace_id_t trid,
					  trace_event_id_t *event,
					  int *unavailable) {

  trace_info_t  *stream;

  // Note:  "stream->current_event" points to the next (not yet reported)
  //        event identifier to be retrieved. Therefore, this value is 
  //        increased right when reporting an event id.

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();

  // Look for the next valid event identifier 
  while (stream->current_event < TRACE_EVENT_MAX && 
	 ! TRACE_EVENT_IS_DEFINED(stream->event_types[stream->current_event])) {
    stream->current_event++;
  }

  if (stream->current_event >= TRACE_EVENT_MAX) {
    // No more valid event ids in the list
    *unavailable = 1;
  } else {
    // Report the event id and increase the counter
    *unavailable = 0;
    *event = stream->event_types[stream->current_event].eventid;
    stream->current_event++;
  }
  return 0;
}





int posix_trace_eventtypelist_rewind (trace_id_t trid) {

  trace_info_t  *stream;
  rtl_irqstate_t flags;  

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();

  // Reset the index of event identifiers to be retrieved
  rtl_no_interrupts(flags);
  stream->current_event = 0;
  rtl_restore_interrupts(flags);
  return 0;
}






//
// 24.3.10. Manipulate trace event type sets
//

// This is implicitly required at least to maintain a void filter in
// each stream, and then be able to build a global filter. Thus, it is
// required even if the _POSIX_TRACE_EVENT_FILTER is not supported
int posix_trace_eventset_empty (trace_event_set_t *set) {
  set->user_set = 0;
  set->system_set = 0;
  return 0;
}



#ifdef _POSIX_TRACE_EVENT_FILTER

int posix_trace_eventset_fill (trace_event_set_t *set,
			       int what) {
  // Note: This implementation consider all system events as
  // process-independent events, since there's only one "process" in
  // the rt-linux execution model (minimal real-time system profile).
  // As a result, WOPID event set is empty.

  // Error checking
  if (what != POSIX_TRACE_WOPID_EVENTS  &&
      what != POSIX_TRACE_SYSTEM_EVENTS &&
      what != POSIX_TRACE_ALL_EVENTS) {
    return -EINVAL;
  }

  set->user_set = 0;
  set->system_set = 0;
 
  if (what == POSIX_TRACE_ALL_EVENTS) {
    set->user_set   = ~0;
    set->system_set = ~0;
  } else if (what == POSIX_TRACE_SYSTEM_EVENTS) {
    set->system_set = ~0;
  }
  return 0;
}






int posix_trace_eventset_add (trace_event_id_t event_id,
			      trace_event_set_t *set) {

  // Error checking
  if (! TRACE_EVENT_IS_VALID(event_id)) {
    return -EINVAL;
  }

  if (event_id < TRACE_USER_EVENT_MAX) {
    set_bit(event_id, &(set->user_set));
  } else {
    set_bit((event_id-32), &(set->system_set));
  }

  return 0;
}






int posix_trace_eventset_del (trace_event_id_t event_id,
			      trace_event_set_t *set) {

  // Error checking
  if (! TRACE_EVENT_IS_VALID(event_id)) {
    return -EINVAL;
  }

  if (event_id < TRACE_USER_EVENT_MAX) {
    clear_bit(event_id, &(set->user_set));
  } else {
    clear_bit((event_id-32), &(set->system_set));
  }

  return 0;
}





int posix_trace_eventset_ismember (trace_event_id_t event_id,
				   const trace_event_set_t *set,
				   int *ismember) {

  // Error checking
  if (! TRACE_EVENT_IS_VALID(event_id)) {
    return -EINVAL;
  }
  
  *ismember = (event_id < TRACE_USER_EVENT_MAX?
	       constant_test_bit(event_id, 
				 (volatile void *) &(set->user_set)) :
	       constant_test_bit((event_id-32),
				  (volatile void *) &(set->system_set)));
  return 0;
}





//
// 24.3.11. Manipulate trace event type sets
//





int posix_trace_set_filter (trace_id_t trid, 
			    const trace_event_set_t *set,
			    int how) {

  trace_info_t     *stream;
  rtl_irqstate_t    flags;  
  struct posix_trace_filter_event_data data;

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  if (how != POSIX_TRACE_SET_EVENTSET &&
      how != POSIX_TRACE_ADD_EVENTSET &&
      how != POSIX_TRACE_SUB_EVENTSET)  {
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();

  // If the stream is running, remember the current filter:
  if (TRACE_STREAM_IS_RUNNING(stream)) {
    data.old_event_filter = stream->filter;
  }

  // Now change the stream filter:
  rtl_no_interrupts(flags);
  switch (how) {
  
  case POSIX_TRACE_SET_EVENTSET :
    stream->filter            = *set;
    break;
  case POSIX_TRACE_ADD_EVENTSET :
    stream->filter.user_set   |= set->user_set;
    stream->filter.system_set |= set->system_set;
    break;
  case POSIX_TRACE_SUB_EVENTSET :
    stream->filter.user_set   &= ~set->user_set;
    stream->filter.system_set &= ~set->system_set;
    break;
  default:
  }
  rtl_restore_interrupts(flags);

#ifdef _POSIX_TRACE_EVENT_SETFILTER
  // Now get the new filter and trace the POSIX_TRACE_FILTER event:
  if (TRACE_STREAM_IS_RUNNING(stream) 
      && ! TRACE_EVENT_IS_FILTERED(POSIX_TRACE_FILTER,stream->filter)) {
    data.new_event_filter = stream->filter;
    stream->do_posix_trace_event(stream, POSIX_TRACE_FILTER, &data, sizeof(data),0);
  }
#endif
  return 0;
}





int posix_trace_get_filter (trace_id_t trid, 
			    trace_event_set_t *set) {

  trace_info_t  *stream;
    
  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();
  *set = stream->filter;

  return 0;
}

#endif // #ifdef _POSIX_TRACE_EVENT_FILTER





//
// 24.3.12. Trace start and stop
//

static inline int _do_posix_trace_start (trace_info_t *stream) {
  rtl_irqstate_t flags;  

  if (! TRACE_STREAM_IS_RUNNING(stream) && ! TRACE_STREAM_IS_FULL(stream)) {
    rtl_no_interrupts(flags);
    stream->status.posix_stream_status = POSIX_TRACE_RUNNING;
    // Remove the set of system events not filtered in this stream:
    global_system_events_filter &= stream->filter.system_set;
    rtl_restore_interrupts(flags);
#ifdef _POSIX_TRACE_EVENT_START    
    if (! TRACE_EVENT_IS_FILTERED(POSIX_TRACE_START,stream->filter)) {   
#ifdef _POSIX_TRACE_EVENT_FILTER
      stream->do_posix_trace_event(stream, POSIX_TRACE_START, 
				   &stream->filter, sizeof(stream->filter), 0);
#else
      stream->do_posix_trace_event(stream, POSIX_TRACE_START, NULL, 0, 0);
#endif
    }
#endif
  }
  return 0;
}




int posix_trace_start (trace_id_t trid) { 

  trace_info_t  *stream;
    
  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  return _do_posix_trace_start(stream);
}





int posix_trace_stop (trace_id_t trid) {

  trace_info_t  *stream;
  rtl_irqstate_t flags;  
#ifdef _POSIX_TRACE_EVENT_STOP
  int            automatic = 0;
#endif

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  if (TRACE_STREAM_IS_RUNNING(stream)
     && ! TRACE_STREAM_IS_FULL(stream)) {

    rtl_no_interrupts(flags);
    stream->status.posix_stream_status = POSIX_TRACE_SUSPENDED;
    // Re-build the global filter of system events:
    build_global_filter();
    rtl_restore_interrupts(flags);
#ifdef _POSIX_TRACE_EVENT_STOP
    if (! TRACE_EVENT_IS_FILTERED(POSIX_TRACE_STOP,stream->filter)) {
      stream->do_posix_trace_event(stream, POSIX_TRACE_STOP, 
				   &automatic, sizeof(int), 0);
    }
#endif
  }
  return 0;
}






//
// 24.3.13. Instrumenting application code
//

#define _do_trace_insert_event(event_write,event_id,data_ptr,data_len, \
			       calling_address,truncated,is_overflow) { \
  event_write->event_id          = event_id; \
  event_write->prog_address      = (void *) calling_address; \
  event_write->truncation_status = truncated; \
  event_write->thread_id         = pthread_self(); \
  rdtscll(event_write->timestamp); \
  event_write->data_length       = (data_len >= 0 ? data_len : 0); \
  event_write->overflow          = is_overflow; \
  if (data_ptr != NULL) { \
    memcpy(&(event_write->user_data), data_ptr, data_len); \
  } \
}





static inline 
 void _do_posix_trace_event_loop (trace_info_t *stream,
				  trace_event_id_t event_id, 
				  const void *data_ptr,      
				  size_t data_len,
				  int calling_address) {
  
  rtl_irqstate_t flags;  
  int            truncation;

  // Note: the checking of the stream running and the event being filtered out
  //      has to be done outside this function.

  truncation  = POSIX_TRACE_NOT_TRUNCATED;
  // If this is a user event, check if we have to truncate its data
  if (data_len > stream->attr.max_data_size && TRACE_EVENT_IS_USER(event_id)) {
    data_len   = stream->attr.max_data_size;
    truncation = POSIX_TRACE_TRUNCATED_RECORD;
  }
  // OK. Let's record the event into the stream.
  rtl_no_interrupts(flags);

  // Note: in LOOP policy, overflow is produced if the stream is
  //       full when tracing an event        
  _do_trace_insert_event(stream->in, event_id, data_ptr, data_len, 
			 calling_address, truncation, 
			 TRACE_STREAM_IS_FULL(stream));
  
  stream->in = TRACE_STREAM_INCREMENT_PTR(stream,stream->in);
  
  if (TRACE_STREAM_IS_FULL(stream)) {
    // If the stream was already full, we have overwritten an event
    stream->out = TRACE_STREAM_INCREMENT_PTR(stream,stream->out);
    stream->status.posix_stream_overrun_status = POSIX_TRACE_OVERRUN;
    
  } else {
    // There's one more event in the stream
    stream->n_events++;
    if (TRACE_STREAM_IS_FULL(stream)) {
      stream->status.posix_stream_full_status = POSIX_TRACE_FULL;
    }
  }

  // Now check if the writing of the event must awake some task:
  if (! TRACE_STACK_IS_EMPTY(stream->wait_stack)) {
    pthread_t      task;
    TRACE_STACK_POP(stream->wait_stack,task);
    if (task) {
      pthread_kill(task, RTL_SIGNAL_WAKEUP);
    }
  } 

  rtl_restore_interrupts(flags);
}





static inline 
 void _do_posix_trace_event_untilfull (trace_info_t *stream,
				       trace_event_id_t event_id, 
				       const void *data_ptr,      
				       size_t data_len,
				       int calling_address) {
  
  rtl_irqstate_t flags;  
  int            truncation;
  int            written;
  
  // Note: the checking of the stream running and the event being filtered out
  //      has to be done outside this function.

  truncation  = POSIX_TRACE_NOT_TRUNCATED;
  written     = 0;
  // If this is a user event, check if we have to truncate its data
  if (data_len > stream->attr.max_data_size && TRACE_EVENT_IS_USER(event_id)) {
    data_len   = stream->attr.max_data_size;
    truncation = POSIX_TRACE_TRUNCATED_RECORD;
  }

  // OK. Let's record the event into the stream.
  rtl_no_interrupts(flags);

  
  if (TRACE_STREAM_IS_FULL(stream)) {
    // In this policy, if the stream is full, we get an overrun, since
    // the event is lost (not recorded in the stream).
    stream->status.posix_stream_overrun_status = POSIX_TRACE_OVERRUN;
  } else {
    if (stream->n_events == stream->max_events - 1) {
#ifdef _POSIX_TRACE_EVENT_STOP
      // If there is room for one event only, record POSIX_TRACE_STOP 
      // and stop the stream (automatically)
      int automatic = 1;
      event_id = POSIX_TRACE_STOP;
      _do_trace_insert_event(stream->in, event_id, &automatic, sizeof(int), 
			     0, POSIX_TRACE_NOT_TRUNCATED, NO_OVERFLOW);
#else
      _do_trace_insert_event(stream->in, event_id, data_ptr, data_len, 
			     calling_address, truncation, NO_OVERFLOW);
#endif
      stream->status.posix_stream_overrun_status = POSIX_TRACE_OVERRUN;
      stream->status.posix_stream_full_status = POSIX_TRACE_FULL;
      stream->status.posix_stream_status      = POSIX_TRACE_SUSPENDED;
      // Re-build the global filter of system events:
      build_global_filter();
    } else {
      // Else, insert the event
      _do_trace_insert_event(stream->in, event_id, data_ptr, data_len, 
			     calling_address, truncation, NO_OVERFLOW);
    }
    
    // In either case, set the stream pointers
    stream->in = TRACE_STREAM_INCREMENT_PTR(stream,stream->in);
    stream->n_events++;
    written = 1;
  }
  
  // Now check if the writing of the event must awake some task:
  if (! written) {
    rtl_restore_interrupts(flags);
    return;
  }

  if (! TRACE_STACK_IS_EMPTY(stream->wait_stack)) {
    pthread_t      task;
    TRACE_STACK_POP(stream->wait_stack,task);
    if (task) {
      pthread_kill(task, RTL_SIGNAL_WAKEUP);
    }
    rtl_restore_interrupts(flags);
    return;
  } 

  rtl_restore_interrupts(flags);
}







#ifdef _POSIX_TRACE_UEVENTS
int posix_trace_eventid_open (const char *event_name, 
			      trace_event_id_t *event_id) {

  // "rtl_trace_event_types" is the array of event types defined
  // for the (only) target process (the rt-linux application)
  return _do_posix_trace_eventid_open(rtl_trace_event_types, 
				      event_name, 
				      event_id);
}





void posix_trace_event (trace_event_id_t event_id, 
			const void *data_ptr,      
			size_t data_len) {

  int           calling_address;
  trace_info_t  *stream;


  if (TRACE_EVENT_IS_VALID(event_id) 
      && TRACE_EVENT_IS_DEFINED(rtl_trace_event_types[event_id])) {

    // Get the invoking address and store it in "calling_address"
    posix_trace_get_invoking_address(calling_address);

#ifdef _POSIX_TRACE_SINGLESTREAM
    if ((stream = rtl_active_stream_list) != EOL
	&& TRACE_STREAM_IS_RUNNING(stream) 
#ifdef _POSIX_TRACE_EVENT_FILTER
	&& ! TRACE_USER_EVENT_IS_FILTERED(event_id,stream->filter)
#endif
	) {
      stream->do_posix_trace_event(stream, event_id, data_ptr, 
				   data_len, calling_address);
    }
#else
    // Trace the event in each stream:
    for (stream = rtl_active_stream_list;
	 stream != EOL;
	 stream = stream->next) {
      if (TRACE_STREAM_IS_RUNNING(stream) 
#ifdef _POSIX_TRACE_EVENT_FILTER
	  && ! TRACE_USER_EVENT_IS_FILTERED(event_id,stream->filter)
#endif
	  ) {
	stream->do_posix_trace_event(stream, event_id, data_ptr, 
				     data_len, calling_address);
      }
    }
#endif
  }
}
#endif




// This next function is called when the call of "posix_trace_event"
// was not done directly to that function, but through some other.
// (and thus the original function call address cannot be retrieved
// from here)
// This is the case of the "posix_trace_event" being called from Linux

void posix_trace_event_with_address (trace_event_id_t event_id, 
				     const void *data_ptr,      
				     size_t data_len,
				     int    calling_address) {

  trace_info_t  *stream;

  if (TRACE_EVENT_IS_VALID(event_id) 
      && TRACE_EVENT_IS_DEFINED(rtl_trace_event_types[event_id])) {

#ifdef _POSIX_TRACE_SINGLESTREAM
    if ((stream = rtl_active_stream_list) != EOL
	&& TRACE_STREAM_IS_RUNNING(stream) 
#ifdef _POSIX_TRACE_EVENT_FILTER
	&& ! TRACE_USER_EVENT_IS_FILTERED(event_id,stream->filter)
#endif
	) {
      stream->do_posix_trace_event(stream, event_id, data_ptr, 
				   data_len, calling_address);
    }
#else
    // Trace the event in each stream:
    for (stream = rtl_active_stream_list;
	 stream != EOL;
	 stream = stream->next) {
      if (TRACE_STREAM_IS_RUNNING(stream) 
#ifdef _POSIX_TRACE_EVENT_FILTER
	  && ! TRACE_EVENT_IS_FILTERED(event_id,stream->filter)
#endif
	  ) {
	stream->do_posix_trace_event(stream, event_id, data_ptr, 
				     data_len, calling_address);
      }
    }
#endif
  }
}





// This next function is used within the RTL modules to report a
// SYSTEM event. Error checking can be omited and the address is also
// given as a parameter (this is needed when we trace a system event 
// corresponding to some system call, in which case the event has to 
// report the address of the syscall invocation)

void posix_trace_system_event (trace_event_id_t event_id, 
			       const void *data_ptr,      
			       size_t data_len,
			       int    calling_address) {
  trace_info_t  *stream;

#ifdef _POSIX_TRACE_SINGLESTREAM
    if ((stream = rtl_active_stream_list) != EOL
	&& TRACE_STREAM_IS_RUNNING(stream) 
#ifdef _POSIX_TRACE_EVENT_FILTER
	&& ! TRACE_USER_EVENT_IS_FILTERED(event_id,stream->filter)
#endif
	) {
	stream->do_posix_trace_event(stream, event_id, data_ptr, 
				     data_len, calling_address);
    }
#else
  for (stream = rtl_active_stream_list;
       stream != EOL;
       stream = stream->next) {
    if (TRACE_STREAM_IS_RUNNING(stream) 
#ifdef _POSIX_TRACE_EVENT_FILTER
	&& ! TRACE_EVENT_IS_FILTERED(event_id,stream->filter)
#endif
	) {
      stream->do_posix_trace_event(stream, event_id, data_ptr, 
				   data_len, calling_address);
    }
  }
#endif
}






#ifdef _POSIX_TRACE_LOGRET
//
// 24.3.14. Trace log management
//

int posix_trace_open (int file_desc, trace_id_t *trid) {
  // Not supported in RT-Linux (only in Linux)
  return -EINVAL;
}

int posix_trace_rewind (trace_id_t trid) { 
  // Not supported in RT-Linux (only in Linux)
  return -EINVAL;
}


int posix_trace_close (trace_id_t trid) {
  // Not supported in RT-Linux (only in Linux)
  return -EINVAL;
}

#endif






//
// 24.3.15. Retrieve the trace attributes or trace statuses
//
#ifdef _POSIX_TRACE_ONLINE
int posix_trace_get_attr (trace_id_t trid,
			  trace_attr_t *attr) {

  trace_info_t  *stream;
    
  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();

  memcpy((void *) attr, (const void *) &(stream->attr), sizeof(trace_attr_t));
  return 0;
}





int posix_trace_get_status (trace_id_t trid,
			    struct posix_trace_status_info *statusinfo) {

  trace_info_t  *stream;
    
  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  // Cancellation test
  pthread_testcancel();

#ifdef _POSIX_TRACE_TOLOG
  stream->status.posix_log_overrun_status = 
    stream->status.posix_stream_overrun_status;
#endif

  memcpy((void *) statusinfo, 
	 (const void *) &(stream->status),
	 sizeof(posix_trace_status_info));

  stream->status.posix_stream_overrun_status = POSIX_TRACE_NO_OVERRUN;
#ifdef _POSIX_TRACE_TOLOG
  stream->status.posix_log_overrun_status = POSIX_TRACE_NO_OVERRUN;
  stream->status.posix_stream_flush_error = 0;
#endif
  return 0;
}





//
// 24.3.16. Retrieve a trace event
//
static inline 
 void _do_trace_retrieve_event (stream_event_t *event_read,
				struct posix_trace_event_info *event,
				void   *data,
				size_t num_bytes,
				size_t *data_len) {

  // Note: this function is called with interrupts disabled
  hrtime_t timestamp_nsecs = tsc2nsecs(event_read->timestamp);

  // Build the event info struct:
  event->posix_event_id          = event_read->event_id;
  event->posix_pid               = 0;
  event->posix_prog_address      = event_read->prog_address;
  event->posix_truncation_status = (num_bytes < event_read->data_length? 
                                    POSIX_TRACE_TRUNCATED_READ :
                                    event_read->truncation_status);
  event->posix_thread_id         = event_read->thread_id;
  event->posix_timestamp         = timespec_from_ns(timestamp_nsecs);

  // Find out how many bytes to copy:
  *data_len = MIN(num_bytes, event_read->data_length);

  // And then copy the event data:
  memcpy((void *) data, (const void *) &(event_read->user_data), *data_len);
}





static inline 
 void _do_posix_trace_getnext_event (trace_info_t  *stream,
				     struct posix_trace_event_info *event,
				     void *data,
				     size_t num_bytes,
				     size_t *data_len) {

  rtl_irqstate_t flags;  

  // Note:  This function must only be invoked if we know for sure that 
  //        there's at least one event available in the stream.

  rtl_no_interrupts(flags);

  // The logic of retrieving an event is different depending on the stream's 
  // full policy:
  if (stream->attr.stream_full_policy ==  POSIX_TRACE_LOOP) {
#if defined _POSIX_TRACE_EVENT_OVERFLOW && defined _POSIX_TRACE_EVENT_RESUME
    // First test an OVERFLOW / RESUME condition: If we detect the
    // beginning or the end of an overflow condition, we do not report
    // the retrieved event but a POSIX_TRACE_OVERFLOW or
    // POSIX_TRACE_RESUME event, with the same timestamp of the
    // retrieved event. Next reported event will be the actual
    // retrieved event.
    int event_overwritten;
    event_overwritten = stream->out->overflow;
    if (! stream->overflow_condition && event_overwritten) {
      // Start of an overflow condition:
      event->posix_event_id          = POSIX_TRACE_OVERFLOW;
      event->posix_truncation_status = POSIX_TRACE_NOT_TRUNCATED;
      event->posix_timestamp       = timespec_from_ns(stream->out->timestamp);
      data = NULL;
      *data_len = 0;
      stream->overflow_condition = 1;
      rtl_restore_interrupts(flags);
      return;
      
    } else if (stream->overflow_condition && ! event_overwritten) {
      // End of an overflow condition:
      event->posix_event_id          = POSIX_TRACE_RESUME;
      event->posix_truncation_status = POSIX_TRACE_NOT_TRUNCATED;
      event->posix_timestamp       = timespec_from_ns(stream->out->timestamp);
      data = NULL;
      *data_len = 0;
      stream->overflow_condition = 0;
      rtl_restore_interrupts(flags);
      return;
    }
#endif
    // No OVERFLOW/RESUME event. Retrieve the event and set the stream
    // pointers:
    _do_trace_retrieve_event(stream->out, event, data, num_bytes, data_len);
    stream->out = TRACE_STREAM_INCREMENT_PTR(stream,stream->out);
    stream->n_events--;
    
    // After an event has been retrieved, the stream is not full (for sure)
    stream->status.posix_stream_full_status = POSIX_TRACE_NOT_FULL;

  } else {

    _do_trace_retrieve_event(stream->out, event, data, num_bytes, data_len);
    stream->out = TRACE_STREAM_INCREMENT_PTR(stream,stream->out);
    stream->n_events--;

    // If the stream was full (and suspended), we only consider it "not
    // full" (and running) again if it's void again. 
    if (stream->status.posix_stream_full_status == POSIX_TRACE_FULL 
	&& stream->n_events == 0) { 

      stream->status.posix_stream_full_status = POSIX_TRACE_NOT_FULL; 
      _do_posix_trace_start(stream);
    }
  }
  // Place here the code for policy POSIX_TRACE_FLUSH (if ever supported)
  // Don't think so, though.

  rtl_restore_interrupts(flags);
}





int posix_trace_getnext_event (trace_id_t trid,
			       struct posix_trace_event_info *event, 
			       void *data,
			       size_t num_bytes,
			       size_t *data_len,
			       int *unavailable) {

  trace_info_t  *stream;
  pthread_t     task;
    
   *unavailable = 1;
 
  // Cancellation test
  pthread_testcancel();

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  if (stream->type != TRACE_STREAM_WOLOG) {
    return -EINVAL;
  }

  if (TRACE_STREAM_IS_EMPTY(stream)) {
    // Insert the task in the wait stack and suspend it
    task = pthread_self();
    if (task == pthread_linux()) {
      return -EINVAL;
    }
    if (TRACE_STACK_IS_FULL(stream->wait_stack)) {
      return -EINVAL;
    } 
    TRACE_STACK_PUSH(stream->wait_stack,task);
    RTL_MARK_SUSPENDED(task);
    rtl_schedule();
  }
  
  // Ok. If we got suspended, we're awake again. While we were asleep,
  // the stream may have been shutdown. Else, we may have been awakened by
  // a signal, and not because an event is on the stream. Check all this!
  if (TRACE_FIND_STREAM(trid) == NO_STREAM) {
    return -EINVAL;
  }
  if (TRACE_STREAM_IS_EMPTY(stream)) { 
    return -EINTR;
  }

  // If no errors, there's an event right there for us. Catch it!
  _do_posix_trace_getnext_event(stream, event, data, num_bytes, data_len);
  *unavailable = 0;
  return 0;
}





int posix_trace_timedgetnext_event (trace_id_t trid,
				    struct posix_trace_event_info *event,
				    void *data,
				    size_t num_bytes,
				    size_t *data_len,
				    int *unavailable,
				    const struct timespec *abs_timeout) {

  trace_info_t  *stream;
  pthread_t     task;
  rtl_sigset_t  awaking_signal;
  hrtime_t      timeout;

  *unavailable = 1;

  // Cancellation test
  pthread_testcancel();

  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  if (stream->type != TRACE_STREAM_WOLOG) {
    return -EINVAL;
  }

  if (! TRACE_STREAM_IS_EMPTY(stream)) {
    // If an event is available, catch it immediately and exit
    _do_posix_trace_getnext_event (stream, event, data, num_bytes, data_len);
    *unavailable = 0;
    return 0;
  } 
  
  // Else, put the task to sleep until there is an available event or the
  // timeout is produced: 
  task = pthread_self();
  if (task == pthread_linux() || TRACE_STACK_IS_FULL(stream->wait_stack)) {
    return -EINVAL;
  }
  
  // Program the timer to "abs_timeout"
  timeout = timespec_to_ns(abs_timeout);
  timeout = __rtl_fix_timeout_for_clock(CLOCK_REALTIME, timeout);
  __rtl_setup_timeout(task, timeout);
  
  // Suspend the task:
  TRACE_STACK_PUSH(stream->wait_stack,task);
  rtl_sigemptyset(&awaking_signal);
  RTL_MARK_SUSPENDED(task);
  awaking_signal = rtl_schedule(); 
  
  // Ok. We're awake again. It may have been for several reasons: the
  // timeout, the stream being shutdown or actually an available event.
  // Check the possible errors first:
  if (RTL_TIMED_OUT(&awaking_signal)) {
    return -ETIMEDOUT;
  }
  if (TRACE_FIND_STREAM(trid) == NO_STREAM 
      || TRACE_STREAM_IS_EMPTY(stream)) {
    return -EINVAL;
  }
  
  // If no errors, there's an event right there for us. Catch it!
  _do_posix_trace_getnext_event (stream, event, data, num_bytes, data_len);
  *unavailable = 0;
  return 0;
}






int posix_trace_trygetnext_event (trace_id_t trid,
				  struct posix_trace_event_info *event,
				  void *data,
				  size_t num_bytes,
				  size_t *data_len,
				  int *unavailable) {

  trace_info_t  *stream;
    
  *unavailable = 1;
  
  // Error checking
  if ((stream = TRACE_FIND_STREAM(trid)) == NO_STREAM) {
    return -EINVAL;
  }

  if (stream->type != TRACE_STREAM_WOLOG) {
    return -EINVAL;
  }

  if (! TRACE_STREAM_IS_EMPTY(stream)) {
    _do_posix_trace_getnext_event (stream, event, data, num_bytes, data_len);
    *unavailable = 0;
  }
  return 0;
}
#endif


 

#ifdef _POSIX_TRACE_LINUXTRACE
/* 
   RTFIFO HANDLERS
*/

// Note: We use a fifo handler to receive calls from some target
//       Linux processes which all use the same fifo (TRACE_TARGET_FIFO)

int rtf_target_handler(unsigned int fifo) {

  int             nbytes;
  fifo_request_t  request;
  fifo_answer_t   answer;

  // Read the request
  nbytes = rtf_get(TRACE_TARGET_FIFO, &request, sizeof(fifo_request_t));

  // Error checking
  if (nbytes != sizeof(fifo_request_t)
      || request.magic != TRACE_FIFO_MAGIC) {
    return -EINVAL;
  }

  // No errors. Let's see which operation has been requested:
  switch (request.operation) {

  case FIFO_TRACE_EVENTID_OPEN: 
    answer.magic        = TRACE_FIFO_MAGIC;
    answer.return_value = posix_trace_eventid_open(request.rdata.eventname,
						  &(answer.data.eventid));
    rtf_put(fifo, &answer, sizeof(fifo_answer_t));
    break;


  case FIFO_TRACE_EVENT :
    posix_trace_event_with_address(request.rdata.tracevent.eventid,
				   request.rdata.tracevent.data,
				   request.rdata.tracevent.data_len,
				   request.rdata.tracevent.calling_address);
    break;

  default :
  }

  return 0;
}
#endif





/* 
   INITIALIZATION / FINALIZATION of TRACE SYSTEM
*/

int trace_system_init (void) {

  int i;

  // Initialize the streams
  for (i = 0; i < TRACE_SYS_MAX; i++) {
    rtl_trace_streams[i].magic = NO_MAGIC;
    rtl_trace_streams[i].next  = NONEXT;
  }

#ifdef _POSIX_TRACE_LINUXTRACE
  // Open the stream for a possible Linux target process
  if (rtf_create(TRACE_TARGET_FIFO, TRACE_FIFO_SIZE) 
      || rtf_create_handler(TRACE_TARGET_FIFO, rtf_target_handler)) {
    return -ENOMEM;
  }
#endif

  // Calculate the maximum size of the data attached to a system event type:
  max_sysevent_data_size = 
    MAX(sizeof(struct posix_trace_error_event_data),
	(MAX(sizeof(struct posix_trace_stop_event_data),
	    sizeof(struct posix_trace_sysevent_data))));


#ifdef _POSIX_TRACE_EVENT_FILTER
  max_sysevent_data_size = 
    MAX(max_sysevent_data_size,
	(MAX(sizeof(struct posix_trace_start_event_data),
	    sizeof(struct posix_trace_filter_event_data))));
#endif


#ifdef _TRACE_DEBUG
  rtl_printf("Max number of events %d\n", TRACE_EVENT_MAX);
#ifdef _POSIX_TRACE_LINUXTRACE
  rtl_printf("Target Linux fifo = %d\n", TRACE_TARGET_FIFO);
#endif

  rtl_printf("sizeof(event header) = %d\n", (int) (sizeof(stream_event_t) - sizeof(void *)));
  rtl_printf("sizeof(error) = %d\n", sizeof(struct posix_trace_error_event_data));
  rtl_printf("sizeof(stop)  = %d\n", sizeof(struct posix_trace_stop_event_data));
  rtl_printf("sizeof(sysev) = %d\n", sizeof(struct posix_trace_sysevent_data));
#ifdef _POSIX_TRACE_EVENT_FILTER
  rtl_printf("sizeof(start) = %d\n", sizeof(struct posix_trace_start_event_data));
	     rtl_printf("sizeof(filter)= %d\n", sizeof(struct posix_trace_filter_event_data));
#endif
  rtl_printf("MAX System Event size = %d\n", max_sysevent_data_size);


  for (i = 0; i< TRACE_EVENT_MAX; i++) {
    rtl_printf("Event number %d   (%s) \n", 
	       rtl_trace_event_types[i].eventid,
	       rtl_trace_event_types[i].name);
  }

#ifdef _POSIX_TRACE_SINGLESTREAM
  rtl_printf("SINGLE stream available\n");
#else
  rtl_printf("Number of streams available = %d\n", TRACE_SYS_MAX);
#endif

#endif

  return 0;
}





int trace_system_cleanup (void) {

  int i;

  // If some mbuff buffers haven't been deallocated, do so...
  for (i=0; i < TRACE_SYS_MAX; i++) {
    posix_trace_shutdown(i);
  }

#ifdef _POSIX_TRACE_LINUXTRACE
  rtf_destroy(TRACE_TARGET_FIFO);
#endif

  return 0;
}


#endif
